import logging
import mlflow
import numpy as np

from typing import Optional
from mlflow.models.signature import infer_signature, ModelSignature
from mlflow.models.utils import ModelInputExample
from .base_mlflow_client import BaseMLflowClient
from .schema import Schema

"""
ModelInputExample
    mlflow 1.x 버전에서는 from mlflow.models.utils import ModelInputExample
    mlflow 2.x 버전에서는 from mlflow.models.signature import ModelInputExample
"""

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


"""
triton inference server 에서 .pth 모델을 사용할 수 없다.
.pt 로 직렬화하고 mlflow.pytorch.log 를 사용해도 .pth 로 저장된다.
log_artifact 는 다른 메타파일이 저장되지 않고 .pt 파일만 업로드해도 UNAVAILABLE 상태로 사용할 수 없다.
torch 모델을 연동하는건 당장 불가능해보인다.
"""


class TorchClient(BaseMLflowClient):
    def upload(
        self,
        model,
        input_example: Optional[Schema] = None,
        output_example: Optional[Schema] = None,
    ):
        signature = infer_signature(input_example, output_example)
        self._log_tensor(model, input_example)
        path = self._log_model(model=model, model_name=os.environ["MODEL_NAME"],, signature=signature)
        try:
            self._channel.basic_publish(
                exchange=self._uploadModelExchange,
                routing_key=self._uploadModelExchange,
                body=path,
            )
        except Exception as e:
            logger.error("Failed to upload model to RabbitMQ: %s", e)

        """
        _log_tensor:
        onnx.load 로 로드된 모델을 통해 input tensor와 output tensor의 정보를 출력합니다.
        """

    def _log_tensor(self, model, input_example):
        hooks = []

        def register_hook(module):
            def hook(module, input, output):
                print(f"{module.__class__.__name__}:")
                print(f"  Input Shape: {list(input[0].size())}")
                print(f"  Output Shape: {list(output.size())}")
                if hasattr(output, "dtype"):
                    print(f"  Data Type: {output.dtype}")

            # Check if it's a layer with parameters; skip for non-layer modules
            if list(module.children()) == []:  # Leaf module, e.g., Conv2d, Linear
                hooks.append(module.register_forward_hook(hook))

        # Apply the hook to all the layers
        model.apply(register_hook)

        # Dummy forward pass to trigger the hooks
        model(input_example)

        # Remove hooks to clean up
        for hook in hooks:
            hook.remove()

    def _log_model(
        self,
        model,
        model_name: str,
        signature: ModelSignature = None,
        input_example: ModelInputExample = None,
    ) -> str:
        with mlflow.start_run():
            mlflow.pytorch.log_model(
                pytorch_model=model,
                artifact_path=model_name,
                signature=signature,
                input_example=input_example,
            )

            artifact_uri = mlflow.get_artifact_uri()
            model_full_path = f"{artifact_uri}/{model_name}"

            logger.info("Full path of the logged model: %s", model_full_path)
            return model_full_path



from typing import TypedDict, Dict
import numpy as np


class Schema(Dict[str, np.ndarray]):
    def __init__(self, params: Dict[str, np.ndarray]):
        super().__init__(params)


class ModelInputOutput:
    def __init__(self, input_schema: Schema, output_schema: Schema):
        self.inputs = input_schema
        self.outputs = output_schema



import logging
import os
import pika
import onnx

from abc import ABCMeta, abstractmethod
from mlflow.models.signature import ModelSignature
from mlflow.models.utils import ModelInputExample
from typing import Optional
from .schema import Schema

"""
ModelInputExample
    mlflow 1.x 버전에서는 from mlflow.models.utils import ModelInputExample
    mlflow 2.x 버전에서는 from mlflow.models.signature import ModelInputExample
"""

logger = logging.getLogger(__name__)


class BaseMLflowClient(metaclass=ABCMeta):
    def __init__(self):
        self._ONNX_TO_TRITON_DTYPE = {
            onnx.TensorProto.BOOL: "TYPE_BOOL",
            onnx.TensorProto.UINT8: "TYPE_UINT8",
            onnx.TensorProto.UINT16: "TYPE_UINT16",
            onnx.TensorProto.UINT32: "TYPE_UINT32",
            onnx.TensorProto.UINT64: "TYPE_UINT64",
            onnx.TensorProto.INT8: "TYPE_INT8",
            onnx.TensorProto.INT16: "TYPE_INT16",
            onnx.TensorProto.INT32: "TYPE_INT32",
            onnx.TensorProto.INT64: "TYPE_INT64",
            onnx.TensorProto.FLOAT16: "TYPE_FP16",
            onnx.TensorProto.FLOAT: "TYPE_FP32",
            onnx.TensorProto.DOUBLE: "TYPE_FP64",
            onnx.TensorProto.STRING: "TYPE_STRING",
            # Brain floating point (bfloat16) is not directly supported in ONNX, you might need to handle it separately
            # "TYPE_BF16": <corresponding ONNX type if available or a custom handler>
        }
        self._uploadModelExchange = os.environ["RABBIT_MODEL_UPLOAD_TOPIC"]

        self._connection = pika.BlockingConnection(
            pika.URLParameters(os.environ["RABBIT_ENDPOINT_URL"])
        )
        self._channel = self._connection.channel()

    def _get_triton_compatible_type(self, tensor_type):
        return self._ONNX_TO_TRITON_DTYPE.get(tensor_type.elem_type, "UNKNOWN")

    @abstractmethod
    def upload(
        self,
        model,
        input_example: Optional[Schema] = None,
        output_example: Optional[Schema] = None,
    ):
        pass

        """
        _log_tensor:
        onnx.load 로 로드된 모델을 통해 input tensor와 output tensor의 정보를 출력합니다.
        """

    @abstractmethod
    def _log_tensor(self, model):
        pass

    @abstractmethod
    def _log_model(
        self,
        model,
        model_name: str,
        signature: ModelSignature = None,
        input_example: ModelInputExample = None,
    ) -> str:
        pass



import os

# 필요한 환경 변수 목록
required_env_vars = [
    "MLFLOW_S3_ENDPOINT_URL",
    "MLFLOW_TRACKING_URI",
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "RABBIT_ENDPOINT_URL",
    "RABBIT_MODEL_UPLOAD_TOPIC",
    "TRAIN_ID",
    "MODEL_NAME",
]

# 환경 변수 검증
missing_vars = [var for var in required_env_vars if not os.environ.get(var)]

# 누락된 환경 변수가 있다면 에러 발생
if missing_vars:
    error_message = f"다음 환경 변수들이 설정되지 않았습니다: {', '.join(missing_vars)}"
    raise EnvironmentError(error_message)



import logging
import mlflow
import onnx
import json
import os

from typing import Optional
from mlflow.models.signature import infer_signature, ModelSignature
from mlflow.models.utils import ModelInputExample
from .base_mlflow_client import BaseMLflowClient
from .schema import Schema

"""
ModelInputExample
    mlflow 1.x 버전에서는 from mlflow.models.utils import ModelInputExample
    mlflow 2.x 버전에서는 from mlflow.models.signature import ModelInputExample
"""

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class OnnxClient(BaseMLflowClient):
    def upload(
        self,
        model,
        input_example: Optional[Schema] = None,
        output_example: Optional[Schema] = None,
    ):
        onnx.checker.check_model(model)
        self._log_tensor(model)

        signature = None
        if input_example is not None and output_example is not None:
            signature = infer_signature(
                model_input=input_example,
                model_output=output_example,
            )

        path = self._log_model(
            model=model,
            model_name=os.environ["MODEL_NAME"],
            signature=signature,
            input_example=input_example,
        )

        message = json.dumps({"train_id": os.environ["TRAIN_ID"], "full_path": path})

        try:
            self._channel.basic_publish(
                exchange=self._uploadModelExchange,
                routing_key=self._uploadModelExchange,
                body=message,
            )
            logger.info("Model uploaded to RabbitMQ: %s", message)
        except Exception as e:
            logger.error("Failed to upload model to RabbitMQ: %s", e)

    def _log_tensor(self, onnx_model):
        """
        onnx.load 로 로드된 모델을 통해 input tensor와 output tensor의 정보를 출력합니다.
        """
        for input_tensor in onnx_model.graph.input:
            input_name = input_tensor.name
            input_type = self._get_triton_compatible_type(input_tensor.type.tensor_type)
            input_shape = [
                dim.dim_value for dim in input_tensor.type.tensor_type.shape.dim
            ]

            print(f"Input tensor name: {input_name}")
            print(f"Data type: {input_type}")
            print(f"Shape: {input_shape}")

        for output_tensor in onnx_model.graph.output:
            output_name = output_tensor.name
            output_type = self._get_triton_compatible_type(
                output_tensor.type.tensor_type
            )
            output_shape = [
                dim.dim_value for dim in output_tensor.type.tensor_type.shape.dim
            ]

            print(f"Output tensor name: {output_name}")
            print(f"Data type: {output_type}")
            print(f"Shape: {output_shape}")

    def _log_model(
        self,
        model,
        model_name: str,
        signature: ModelSignature = None,
        input_example: ModelInputExample = None,
    ) -> str:
        with mlflow.start_run():
            mlflow.onnx.log_model(
                onnx_model=model,
                artifact_path=model_name,
                signature=signature,
                input_example=input_example,
            )

            artifact_uri = mlflow.get_artifact_uri()
            model_full_path = f"{artifact_uri}/{model_name}"

            logger.info("Full path of the logged model: %s", model_full_path)
            return model_full_path



#!/usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name="wtu-mlflow",
    version="0.0.9",
    author="hbjs",
    author_email="hbjs97@naver.com",
    description="W-Train Utils for MLflow",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/WIM-Corporation/w-train-utils-mlflow",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.7",
    packages=find_packages(exclude=("tests", "tests.*")),
    install_requires=[
        # "mlflow>=1.30.1,<3.0",
        "mlflow==1.30.1",
        "numpy>=1.21.6",
        "boto3>=1.24.0",
        "pika>=0.13.0",
        "onnx",
        "onnxruntime",
    ],
)
